@@ -4,7 +4,7 @@
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation, either version 2.1 of the License, or
+ * the Free Software Foundation, either version 2 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -30,6 +30,7 @@
 #include "shortcuts/dzl-shortcuts-group.h"
 #include "shortcuts/dzl-shortcuts-section.h"
 #include "shortcuts/dzl-shortcuts-shortcut.h"
+#include "util/dzl-gtk.h"
 #include "util/dzl-util-private.h"
 
 typedef struct
@@ -408,12 +409,40 @@ dzl_shortcut_manager_class_init (DzlShortcutManagerClass *klass)
                   0, NULL, NULL, NULL, G_TYPE_NONE, 0);
 }
 
+static guint
+shortcut_entry_hash (gconstpointer key)
+{
+  DzlShortcutEntry *entry = (DzlShortcutEntry *)key;
+  guint command_hash = 0;
+  guint section_hash = 0;
+  guint group_hash = 0;
+  guint title_hash = 0;
+  guint subtitle_hash = 0;
+
+  if (entry->command != NULL)
+    command_hash = g_str_hash (entry->command);
+
+  if (entry->section != NULL)
+    section_hash = g_str_hash (entry->section);
+
+  if (entry->group != NULL)
+    group_hash = g_str_hash (entry->group);
+
+  if (entry->title != NULL)
+    title_hash = g_str_hash (entry->title);
+
+  if (entry->subtitle != NULL)
+    subtitle_hash = g_str_hash (entry->subtitle);
+
+  return (command_hash ^ section_hash ^ group_hash ^ title_hash ^ subtitle_hash);
+}
+
 static void
 dzl_shortcut_manager_init (DzlShortcutManager *self)
 {
   DzlShortcutManagerPrivate *priv = dzl_shortcut_manager_get_instance_private (self);
 
-  priv->seen_entries = g_hash_table_new (NULL, NULL);
+  priv->seen_entries = g_hash_table_new (shortcut_entry_hash, NULL);
   priv->themes = g_ptr_array_new_with_free_func (destroy_theme);
   priv->root = g_node_new (NULL);
   priv->internal_theme = g_object_new (DZL_TYPE_SHORTCUT_THEME,
@@ -647,7 +676,6 @@ dzl_shortcut_manager_set_theme (DzlShortcutManager *self,
  * @phase: the phase (capture, bubble)
  * @widget: the widget the event was destined for
  * @focus: the current focus widget
- * @toplevel: the toplevel of the event window
  *
  * Runs a particular phase of the event dispatch.
  *
@@ -670,7 +698,6 @@ dzl_shortcut_manager_run_phase (DzlShortcutManager     *self,
 {
   GtkWidget *ancestor = widget;
   GQueue queue = G_QUEUE_INIT;
-  GdkModifierType modifier;
   DzlShortcutMatch ret = DZL_SHORTCUT_MATCH_NONE;
 
   g_assert (DZL_IS_SHORTCUT_MANAGER (self));
@@ -680,8 +707,6 @@ dzl_shortcut_manager_run_phase (DzlShortcutManager     *self,
   g_assert (GTK_IS_WIDGET (widget));
   g_assert (GTK_IS_WIDGET (focus));
 
-  modifier = event->state & gtk_accelerator_get_default_mod_mask ();
-
   /*
    * Collect all the widgets that might be needed for this phase and order them
    * so that we can process from first-to-last. Capture phase is
@@ -705,29 +730,11 @@ dzl_shortcut_manager_run_phase (DzlShortcutManager     *self,
     {
       GtkWidget *current = iter->data;
       DzlShortcutController *controller;
-      gboolean use_binding_sets = TRUE;
 
       controller = dzl_shortcut_controller_try_find (current);
 
       if (controller != NULL)
         {
-          if (phase == DZL_SHORTCUT_PHASE_DISPATCH)
-            {
-              DzlShortcutContext *context;
-
-              /*
-               * Get the default context for the controller. We do not take
-               * the controller for this particular phase because we only are
-               * using the context to determine if we should dispatch the
-               * default binding sets later on.
-               */
-              context = dzl_shortcut_controller_get_context (controller);
-              if (context != NULL)
-                g_object_get (context,
-                              "use-binding-sets", &use_binding_sets,
-                              NULL);
-            }
-
           /*
            * Now try to activate the event using the controller. If we get
            * any result other than DZL_SHORTCUT_MATCH_NONE, we need to stop
@@ -743,48 +750,16 @@ dzl_shortcut_manager_run_phase (DzlShortcutManager     *self,
         }
 
       /*
-       * If the current context at activation indicates that we can
-       * dispatch using the default binding sets for the widget, go
-       * ahead and try to do that.
+       * If we are in the dispatch phase, we will only see our target widget for
+       * the event delivery. Try to dispatch the event and if so we consider
+       * the event handled.
        */
-      if (use_binding_sets && phase == DZL_SHORTCUT_PHASE_DISPATCH)
+      if (phase == DZL_SHORTCUT_PHASE_DISPATCH)
         {
-          GtkStyleContext *style_context;
-          g_autoptr(GPtrArray) sets = NULL;
-
-          style_context = gtk_widget_get_style_context (current);
-          gtk_style_context_get (style_context,
-                                 gtk_style_context_get_state (style_context),
-                                 "-gtk-key-bindings", &sets,
-                                 NULL);
-
-          if (sets != NULL)
+          if (gtk_widget_event (current, (GdkEvent *)event))
             {
-              for (guint i = 0; i < sets->len; i++)
-                {
-                  GtkBindingSet *set = g_ptr_array_index (sets, i);
-
-                  if (gtk_binding_set_activate (set, event->keyval, modifier, G_OBJECT (current)))
-                    {
-                      ret = DZL_SHORTCUT_MATCH_EQUAL;
-                      DZL_GOTO (cleanup);
-                    }
-                }
-            }
-
-          /*
-           * Only if this widget is also our focus, try to activate the default
-           * keybindings for the widget.
-           */
-          if (current == focus)
-            {
-              GtkBindingSet *set = gtk_binding_set_by_class (G_OBJECT_GET_CLASS (current));
-
-              if (gtk_binding_set_activate (set, event->keyval, modifier, G_OBJECT (current)))
-                {
-                  ret = DZL_SHORTCUT_MATCH_EQUAL;
-                  DZL_GOTO (cleanup);
-                }
+              ret = DZL_SHORTCUT_MATCH_EQUAL;
+              DZL_GOTO (cleanup);
             }
         }
     }
@@ -823,6 +798,51 @@ dzl_shortcut_manager_run_global (DzlShortcutManager     *self,
   return _dzl_shortcut_controller_handle (root, event, chord, phase, widget);
 }
 
+static gboolean
+dzl_shortcut_manager_run_fallbacks (DzlShortcutManager     *self,
+                                    GtkWidget              *widget,
+                                    const DzlShortcutChord *chord)
+{
+  g_assert (DZL_IS_SHORTCUT_MANAGER (self));
+  g_assert (GTK_IS_WIDGET (widget));
+  g_assert (chord != NULL);
+
+  if (dzl_shortcut_chord_get_length (chord) == 1)
+    {
+      GApplication *app = g_application_get_default ();
+
+      if (GTK_IS_APPLICATION (app))
+        {
+          g_autofree gchar *accel = dzl_shortcut_chord_to_string (chord);
+          g_auto(GStrv) actions = NULL;
+
+          actions = gtk_application_get_actions_for_accel (GTK_APPLICATION (app), accel);
+
+          if (actions != NULL)
+            {
+              for (guint i = 0; actions[i] != NULL; i++)
+                {
+                  const gchar *action = actions[i];
+                  g_autofree gchar *prefix = NULL;
+                  g_autofree gchar *name = NULL;
+                  g_autoptr(GVariant) param = NULL;
+
+                  if (!dzl_g_action_name_parse_full (action, &prefix, &name, &param))
+                    {
+                      g_warning ("Failed to parse: %s", action);
+                      continue;
+                    }
+
+                  if (dzl_gtk_widget_action (widget, prefix, name, param))
+                    return TRUE;
+                }
+            }
+        }
+    }
+
+  return FALSE;
+}
+
 /**
  * dzl_shortcut_manager_handle_event:
  * @self: (nullable): An #DzlShortcutManager
@@ -909,7 +929,8 @@ dzl_shortcut_manager_handle_event (DzlShortcutManager *self,
       (match = dzl_shortcut_manager_run_phase (self, event, chord, DZL_SHORTCUT_PHASE_CAPTURE, widget, focus)) ||
       (match = dzl_shortcut_manager_run_phase (self, event, chord, DZL_SHORTCUT_PHASE_DISPATCH, widget, focus)) ||
       (match = dzl_shortcut_manager_run_phase (self, event, chord, DZL_SHORTCUT_PHASE_BUBBLE, widget, focus)) ||
-      (match = dzl_shortcut_manager_run_global (self, event, chord, DZL_SHORTCUT_PHASE_BUBBLE, root, widget)))
+      (match = dzl_shortcut_manager_run_global (self, event, chord, DZL_SHORTCUT_PHASE_BUBBLE, root, widget)) ||
+      (match = dzl_shortcut_manager_run_fallbacks (self, widget, chord)))
     ret = GDK_EVENT_STOP;
 
   DZL_TRACE_MSG ("match = %d", match);
@@ -1351,7 +1372,6 @@ dzl_shortcut_manager_add_shortcuts_to_window (DzlShortcutManager *self,
             {
               DzlShortcutNodeData *data = iter->data;
               const DzlShortcutChord *chord = NULL;
-              g_autofree gchar *accel = NULL;
               DzlShortcutsShortcut *shortcut;
 
               if (data->type == DZL_SHORTCUT_NODE_ACTION)
@@ -1359,8 +1379,6 @@ dzl_shortcut_manager_add_shortcuts_to_window (DzlShortcutManager *self,
               else if (data->type == DZL_SHORTCUT_NODE_COMMAND)
                 chord = dzl_shortcut_theme_get_chord_for_command (theme, data->name);
 
-              accel = dzl_shortcut_chord_to_string (chord);
-
               shortcut = create_shortcut (chord, data->title, data->subtitle);
               gtk_container_add (GTK_CONTAINER (group), GTK_WIDGET (shortcut));
             }
@@ -1554,7 +1572,7 @@ dzl_shortcut_manager_merge (DzlShortcutManager *self,
    * Okay, now we need to go through all the custom contexts, and global
    * shortcuts in the theme and merge them into the base_layer. However, we
    * will defer that work to the DzlShortcutTheme module so it has access to
-   * the intenral structures.
+   * the internal structures.
    */
   _dzl_shortcut_theme_merge (base_layer, theme);
 
